# -*- coding: utf-8 -*-
"""Control parameters for a MOPAC single-point energy calculation
"""

import logging
import seamm

logger = logging.getLogger(__name__)


class EnergyParameters(seamm.Parameters):
    """The control parameters for creating a structure from SMILES"""

    parameters = {
        "structure": {
            "default": "default",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "default",
                "initial",
                "current",
            ),
            "format_string": "s",
            "description": "Structure:",
            "help_text": (
                "The structure to use. By default, for the "
                "first calculation the incoming structure is "
                "used, and subsequently that from the previous "
                "MOPAC steps."
            ),
        },
        "hamiltonian": {
            "default": "PM7",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "AM1",
                "MNDO",
                "MNDOD",
                "PM3",
                "PM6",
                "PM6-D3",
                "PM6-DH+",
                "PM6-DH2",
                "PM6-DH2X",
                "PM6-D3H4",
                "PM6-D3H4X",
                "PM7",
                "PM7-TS",
                "RM1",
            ),
            "format_string": "s",
            "description": "Hamiltonian:",
            "help_text": ("The Hamiltonian (parameterization) to use."),
        },
        "convergence": {
            "default": "normal",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "normal",
                "precise",
                "relative",
                "absolute",
            ),
            "format_string": "s",
            "description": "Convergence criterion:",
            "help_text": (
                "The convergence criterion for the energy. "
                "* 'normal' is the default level, 1.0E-04 "
                " kcal/mol for SCF. For force contants, transition "
                " state searches, etc. the default is 1.0E-07."
                "* 'precise' tightens up the criterion by a factor "
                "of 100."
                "* 'relative' multiplies the SCF convergence "
                "criterion by the amount given. Values >1 loosen "
                "the tolerance, while values <1 tighten it."
                "* 'absolute' sets the criterion directly."
            ),
        },
        "uhf": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": (
                "yes",
                "no",
            ),
            "format_string": "s",
            "description": "UHF for singlets:",
            "help_text": "Whether to use UHF for singlet states.",
        },
        "COSMO": {
            "default": "no",
            "kind": "boolean",
            "default_units": "",
            "enumeration": (
                "yes",
                "no",
            ),
            "format_string": "s",
            "description": "Solvent using COSMO:",
            "help_text": "Whether to use COSMO solvation model.",
        },
        "MOZYME": {
            "default": "for larger systems",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "for larger systems",
                "always",
                "never",
            ),
            "format_string": "s",
            "description": "Use localized molecular orbitals (MOZYME):",
            "help_text": (
                "Whether and when to use localized molecular orbitals (LMO) by using "
                "MOZYME."
            ),
        },
        "nMOZYME": {
            "default": 300,
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "with this many atoms or more:",
            "help_text": ("The number of atoms to switch to using MOZYME."),
        },
        "MOZYME follow-up": {
            "default": (
                "recalculate the energy at the end using new, orthogonal localized "
                "orbitals"
            ),
            "kind": "enumeration",
            "default_units": "",
            "enumeration": (
                "recalculate the energy at the end using new, orthogonal localized "
                "orbitals",
                "recalculate the energy at the end using exact, non-localized orbitals",
                "none",
            ),
            "format_string": "s",
            "description": "Follow-up calculation:",
            "help_text": (
                "Whether to follow the localize molecular orbital calculation "
                "with another calculation to clean up the orbitals, or check "
                "by doing a full calculation without localizing the orbitals."
            ),
        },
        "eps": {
            "default": "78.4",
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "Dielectric constant:",
            "help_text": "The solvent's dielectric constant.",
        },
        "rsolve": {
            "default": "1.3",
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": "",
            "description": "Solvent radius:",
            "help_text": "The solvent's approximate radius.",
        },
        "nspa": {
            "default": 42,
            "kind": "integer",
            "default_units": "",
            "enumeration": (
                "12",
                "32",
                "42",
                "92",
                "122",
                "162",
                "252",
                "272",
                "362",
                "482",
                "492",
                "752",
                "812",
                "1082",
                "1442",
                "1472",
            ),
            "format_string": "",
            "description": "Surface grid size:",
            "help_text": (
                "The number of points in the solvent-accessible surface grid per atom."
            ),
        },
        "disex": {
            "default": "2.0",
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "Cutoff:",
            "help_text": (
                "The cutoff for exact calculation of segment-segment interactions in "
                "COSMO."
            ),
        },
        "calculate gradients": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": (
                "yes",
                "no",
            ),
            "format_string": "s",
            "description": "Calculate gradients:",
            "help_text": "Whether to calculate the gradients.",
        },
        "relative": {
            "default": "0.1",
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "factor",
            "help_text": (
                "The factor to multiply the default SCF "
                "convergence criterion. Values >1 make the "
                "criterion looser; < 1, tighter."
            ),
        },
        "absolute": {
            "default": "1.0E-07",
            "kind": "float",
            "default_units": "kcal/mol",
            "enumeration": tuple(),
            "format_string": "",
            "description": "criterion",
            "help_text": (
                "The SCF convergence criterion, based on the "
                "change in energy between iterations."
            ),
        },
        "extra keywords": {
            "default": [],
            "kind": "list",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "Extra keywords",
            "help_text": (
                "Extra keywords to append to those from the GUI. "
                "This allows you to add to and override the GUI."
            ),
        },
        "results": {
            "default": {},
            "kind": "dictionary",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "",
            "description": "results",
            "help_text": ("The results to save to variables or in " "tables. "),
        },
        "create tables": {
            "default": "yes",
            "kind": "boolean",
            "default_units": "",
            "enumeration": ("yes", "no"),
            "format_string": "",
            "description": "Create tables as needed:",
            "help_text": (
                "Whether to create tables as needed for "
                "results being saved into tables."
            ),
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={**EnergyParameters.parameters, **defaults}, data=data
        )



# -*- coding: utf-8 -*-

"""Run a vibrational frequency calculation in MOPAC"""

import csv
import logging
from pathlib import Path
import textwrap
import traceback

from tabulate import tabulate

import mopac_step
import seamm
import seamm_util.printing as printing
from seamm_util import units_class
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")


class IR(mopac_step.Energy):
    def __init__(self, flowchart=None, title="IR Spectrum", extension=None):
        """Initialize the node"""

        logger.debug("Creating IR {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self.parameters = mopac_step.IRParameters()

        # Customize the options for naming the configuration
        p = self.parameters["configuration name"]
        enumeration = (
            "use SMILES string",
            "use Canonical SMILES string",
            "keep current name",
            "vibrations with <Hamiltonian>",
            "use configuration number",
        )
        p.default = "vibrations with <Hamiltonian>"
        if p.value in p._data["enumeration"] and p.value not in enumeration:
            p.value = p.default
        p._data["enumeration"] = enumeration

        self.description = "Infrared (vibrational) spectroscopy calculation"

    def description_text(self, P=None):
        """Prepare information about what this node will do"""

        if not P:
            P = self.parameters.values_to_dict()

        # The energy part of the description
        tmp = super().description_text(P)
        energy_description = textwrap.dedent("\n".join(tmp.splitlines()[1:]))

        text = (
            "Harmonic vibrational calculation using {hamiltonian}, "
            "with the SCF converged to "
        )
        # Convergence
        if P["convergence"] == "normal":
            text += "the 'normal' level of 1.0e-07 kcal/mol."
        elif P["convergence"] == "precise":
            text += "the 'precise' level of 1.0e-09 kcal/mol."
        elif P["convergence"] == "relative":
            text += "a factor of {relative} times the " "normal criterion."
        elif P["convergence"] == "absolute":
            text += "converged to {absolute}."

        # Put in the description of the energy calculation
        text += "\n\nThe energy and forces will be c" + energy_description[1:]
        text += "\n\n"

        # Structure handling
        text += "The structure in the standard orientation will {structure handling} "

        confname = P["configuration name"]
        if confname == "use SMILES string":
            text += "using SMILES as its name."
        elif confname == "use Canonical SMILES string":
            text += "using canonical SMILES as its name."
        elif confname == "keep current name":
            text += "keeping the current name."
        elif confname == "vibrations with <Hamiltonian>":
            text += "with 'vibrations with {hamiltonian}' as its name."
        elif confname == "use configuration number":
            text += "using the index of the configuration (1, 2, ...) as its name."
        else:
            confname = confname.replace("<Hamiltonian>", P["hamiltonian"])
            text += f"with '{confname}' as its name."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self):
        """Get the input for an optimization MOPAC"""

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        # Save the description for later printing
        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=self.indent))

        # Remove the 1SCF keyword from the energy setup
        keywords = []
        for keyword in super().get_input():
            if keyword == "1SCF":
                keywords.append("FORCE")
                if P["let"]:
                    keywords.append("LET")
            else:
                keywords.append(keyword)

        return keywords

    def analyze(self, indent="", data={}, out=[], table=None):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """
        # Update the structure
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        if "ORIENTATION_ATOM_X" in data:
            system, starting_configuration = self.get_system_configuration(None)
            periodicity = starting_configuration.periodicity
            if (
                "structure handling" in P
                and P["structure handling"] == "be put in a new configuration"
            ):
                configuration = system.create_configuration(
                    periodicity=periodicity,
                    atomset=starting_configuration.atomset,
                    bondset=starting_configuration.bondset,
                    cell_id=starting_configuration.cell_id,
                )
                configuration.charge = starting_configuration.charge
                configuration.spin_multiplicity = (
                    starting_configuration.spin_multiplicity
                )
            else:
                configuration = starting_configuration

            if periodicity != 0:
                raise NotImplementedError(
                    "Thermodynamics cannot yet handle periodicity"
                )
            xyz = []
            it = iter(data["ORIENTATION_ATOM_X"])
            for x in it:
                xyz.append([float(x), float(next(it)), float(next(it))])
            configuration.atoms.set_coordinates(xyz, fractionals=False)

            # And the name of the configuration.
            if "configuration name" in P:
                if P["configuration name"] == "use SMILES string":
                    configuration.name = configuration.smiles
                elif P["configuration name"] == "use Canonical SMILES string":
                    configuration.name = configuration.canonical_smiles
                elif P["configuration name"] == "keep current name":
                    pass
                elif P["configuration name"] == "vibrations with <Hamiltonian>":
                    configuration.name = f"vibrations with {P['hamiltonian']}"
                else:
                    confname = P["configuration name"]
                    confname = confname.replace("<Hamiltonian>", P["hamiltonian"])
                    configuration.name = confname

        # Write the structure out for viewing.
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        #  MMCIF file has bonds
        try:
            path = directory / "optimized.mmcif"
            path.write_text(configuration.to_mmcif_text())
        except Exception:
            message = "Error creating the mmcif file\n\n" + traceback.format_exc()
            logger.warning(message)
        # CIF file has cell
        if configuration.periodicity == 3:
            try:
                path = directory / "optimized.cif"
                path.write_text(configuration.to_cif_text())
            except Exception:
                message = "Error creating the cif file\n\n" + traceback.format_exc()
                logger.warning(message)

        # Print the moments of inertia
        p1, p2, p3 = data["PRI_MOM_OF_I"]
        r1, r2, r3 = data["ROTAT_CONSTS"]
        tbl = {
            "Property": ("Principal moment of inertia", "Rotational constants"),
            "1": (p1, r1),
            "2": (p2, r2),
            "3": (p3, r3),
            "Units": ("1.0E-40 g.cm^2", "1/cm"),
        }
        text_lines = "                  Rotational Constants\n"
        text_lines += tabulate(
            tbl,
            headers="keys",
            tablefmt="psql",
            colalign=("center", "decimal", "decimal", "decimal", "left"),
        )
        text_lines += "\n"
        text = textwrap.indent(text_lines, self.indent + 7 * " ")
        printer.normal(text)

        # And the vibrational modes to a csv file
        # First, how many rotations are there?
        n_rot = sum([0 if PMI < 0.001 else 1 for PMI in data["PRI_MOM_OF_I"]])
        n_vib = len(data["VIB._FREQ"]) - 3 - n_rot
        with open(directory / "vibrations.csv", "w", newline="") as fd:
            writer = csv.writer(fd)
            writer.writerow(
                (
                    "Mode",
                    "Frequency (1/cm)",
                    "Symmetry",
                    "Transition Dipole (e-)",
                    "Travel (Å)",
                    "Reduced Mass (amu)",
                    "Effective Mass (amu)",
                )
            )
            for row in zip(
                range(1, n_vib + 1),
                data["VIB._FREQ"][0:n_vib],
                data["NORMAL_MODE_SYMMETRY_LABELS"][0:n_vib],
                data["VIB._T_DIP"][0:n_vib],
                data["VIB._TRAVEL"][0:n_vib],
                data["VIB._RED_MASS"][0:n_vib],
                data["VIB._EFF_MASS"][0:n_vib],
            ):
                writer.writerow(row)

        # Let the energy module do its thing
        super().analyze(indent=indent, data=data, out=out, table=table)



# -*- coding: utf-8 -*-

"""Top-level package for MOPAC step."""

# Bring up the classes so that they appear to be directly in
# the package.

from mopac_step.mopac_step import MOPACStep  # noqa: F401
from mopac_step.mopac import MOPAC  # noqa: F401
from mopac_step.tk_mopac import TkMOPAC  # noqa: F401
from mopac_step.energy_step import EnergyStep  # noqa: F401
from mopac_step.energy import Energy  # noqa: F401
from mopac_step.energy_parameters import EnergyParameters  # noqa: F401
from mopac_step.tk_energy import TkEnergy  # noqa: F401
from mopac_step.optimization_step import OptimizationStep  # noqa: F401
from mopac_step.optimization import Optimization  # noqa: F401
from mopac_step.optimization_parameters import OptimizationParameters  # noqa: F401
from mopac_step.tk_optimization import TkOptimization  # noqa: F401
from mopac_step.ir_step import IRStep  # noqa: F401
from mopac_step.ir import IR  # noqa: F401
from mopac_step.ir_parameters import IRParameters  # noqa: F401
from mopac_step.tk_ir import TkIR  # noqa: F401
from mopac_step.thermodynamics_step import ThermodynamicsStep  # noqa: F401
from mopac_step.thermodynamics import Thermodynamics  # noqa: F401
from mopac_step.thermodynamics_parameters import (  # noqa: F401
    ThermodynamicsParameters,
)
from mopac_step.tk_thermodynamics import TkThermodynamics  # noqa: F401

# Handle versioneer
from ._version import get_versions

# Parameters used for handling the structure if it is changed.
structure_handling_parameters = {
    "structure handling": {
        "default": "be put in a new configuration",
        "kind": "enum",
        "default_units": "",
        "enumeration": (
            "overwrite the current configuration",
            "be put in a new configuration",
        ),
        "format_string": "s",
        "description": "Optimized structure will",
        "help_text": (
            "Whether to overwrite the current configuration, or create a new "
            "configuration or system and configuration for the new structure"
        ),
    },
    "configuration name": {
        "default": "optimized with <Hamiltonian>",
        "kind": "string",
        "default_units": "",
        "enumeration": (
            "optimized with <Hamiltonian>",
            "keep current name",
            "use SMILES string",
            "use Canonical SMILES string",
            "use configuration number",
        ),
        "format_string": "s",
        "description": "Configuration name:",
        "help_text": "The name for the new configuration",
    },
}

__author__ = """Paul Saxe"""
__email__ = "psaxe@molssi.org"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions

keyword_metadata = {
    "0SCF": {
        "description": "Read in data, then stop",
    },
    "1ELECTRON": {
        "description": "Print final one-electron matrix",
    },
    "1SCF": {
        "description": "Do one scf and then stop",
    },
    "ADD-H": {
        "description": (
            "Add hydrogen atoms (intended for use with organic " "compounds)"
        ),
    },
    "A0": {
        "description": "Input geometry is in atomic units",
    },
    "AIDER": {
        "description": "Read in ab-initio derivatives",
    },
    "AIGIN": {
        "description": "Geometry must be in Gaussian format",
    },
    "AIGOUT": {
        "description": "Print the geometry in Gaussian format in the ARC file",
    },
    "ALLBONDS": {
        "description": (
            "Print final bond-order matrix, including bonds to " "hydrogen"
        ),
    },
    "ALLVEC": {
        "description": "Print all vectors (keywords vectors also needed)",
    },
    "ALT_A": {
        "description": "In PDB files with alternative atoms, select atoms A",
        "takes values": 1,
        "default": "A",
        "format": "{}={}",
    },
    "ANGSTROMS": {
        "description": "Input geometry is in Angstroms",
    },
    "AUTOSYM": {
        "description": "Symmetry to be imposed automatically",
    },
    "AUX": {
        "description": ("Output auxiliary information for use by other " "programs"),
    },
    "AM1": {
        "description": "Use the AM1 hamiltonian",
    },
    "BAR": {
        "description": "reduce bar length by a maximum of n.nn%",
        "takes values": 1,
        "default": "0.01",
        "format": "{}={}",
    },
    "BCC": {
        "description": "Only even unit cells used (used by BZ)",
    },
    "BIGCYCLES": {
        "description": "Do a maximum of n big steps",
        "takes values": 1,
        "default": "1",
        "format": "{}={}",
    },
    "BIRADICAL": {
        "description": "System has two unpaired electrons",
    },
    "BFGS": {
        "description": "Use the Flepo or BFGS geometry optimizer",
    },
    "BONDS": {
        "description": "Print final bond-order matrix",
    },
    "CAMP": {
        "description": "Use Camp-King converger in SCF",
    },
    "CARTAB": {
        "description": "Print point-group character table",
    },
    "C.I.": {
        "description": "A multi-electron configuration interaction specified",
        "takes values": [1, 2],
        "default": "2",
        "format": "{}={}",
    },
    "CHAINS(text)": {
        "description": ("In a protein, explicitly define the letters of " "chains."),
    },
    "CHECK": {
        "description": "Report possible faults in input geometry",
    },
    "CHARGE": {
        "description": "Charge on system = n (e.g. NH4 = +1)",
        "takes values": 1,
        "default": "+1",
        "format": "{}={}",
    },
    "CHARGES": {
        "description": (
            "Print net charge on the system, and all charges in " "the system"
        ),
    },
    "CHARST": {
        "description": "Print details of working in CHARST",
    },
    "CIS": {
        "description": "C.I. uses 1 electron excitations only",
    },
    "CISD": {
        "description": "C.I. uses 1 and 2 electron excitations",
    },
    "CISDT": {
        "description": "C.I. uses 1, 2 and 3 electron excitations",
    },
    "COMPARE": {
        "description": "Compare the geometries of two systems",
    },
    "COMPFG": {
        "description": "Print heat of formation calculated in COMPFG",
    },
    "COSCCH": {
        "description": "Add in COSMO charge corrections",
    },
    "COSWRT": {
        "description": ("Write details of the solvent accessible surface to a " "file"),
    },
    "CUTOFP": {
        "description": "Madelung distance cutoff is n .nn Angstroms",
        "takes values": 1,
        "default": "15.0",
        "format": "{}={}",
    },
    "CUTOFF": {
        "description": (
            "In MOZYME, the interatomic distance where the NDDO " "approximation stops"
        ),
        "takes values": 1,
        "default": "6.0",
        "format": "{}={}",
    },
    "CUTOF1": {
        "description": "In MOZYME, the cutoff distance for polarization",
        "takes values": 1,
        "default": "10.0",
        "format": "{}={}",
    },
    "CUTOF2": {
        "description": ("In MOZYME, the cutoff distance for two-center " "integral"),
        "takes values": 1,
        "default": "9.9",
        "format": "{}={}",
    },
    "CUTOFS": {
        "description": "In MOZYME, the cutoff distance for overlap integrals",
        "takes values": 1,
        "default": "9.0",
        "format": "{}={}",
    },
    "CYCLES": {
        "description": "Do a maximum of n steps of geometry optimization",
        "takes values": 1,
        "default": "50",
        "format": "{}={}",
    },
    "CVB": {
        "description": (
            "In MOZYME. add and remove specific bonds to allow a "
            "Lewis or PDB structure."
        ),
    },
    "DAMP": {
        "description": (
            "in MOZYME. damp SCF oscillations using a factor of "
            "n.nn, 0<n.nn<1.0, <0.5 for solids"
        ),
        "takes values": 1,
        "default": "0.5",
        "format": "{}={}",
    },
    "DATA": {
        "description": "Input data set is re-defined to text",
        "takes values": 1,
        "default": "/file.dat",
        "format": "{}={}",
    },
    "DCART": {
        "description": "Print part of working in DCART",
    },
    "DDMAX": {
        "description": (
            "The maximum value of the trust radius in EF and TS. "
            "Defaults to 0.3 in TS and 0.5 in EF"
        ),
        "takes values": 1,
        "default": "0.3",
        "format": "{}={}",
    },
    "DDMIN": {
        "description": (
            "Minimum trust radius in a EF/TS calculation. "
            "Defaults to 0.001. Use smaller values close to "
            "minimium"
        ),
        "takes values": 1,
        "default": "0.00001",
        "format": "{}={}",
    },
    "DEBUG": {
        "description": "Debug option turned on",
    },
    "DEBUG PULAY": {
        "description": "Print working in PULAY",
    },
    "DENOUT": {
        "description": "Density matrix output, unformatted",
    },
    "DENOUTF": {
        "description": "Density matrix output, formatted",
    },
    "DENSITY": {
        "description": "Print final density matrix",
    },
    "DERI1": {
        "description": "Print part of working in DERI1",
    },
    "DERI2": {
        "description": "Print part of working in DERI2",
    },
    "DERITR": {
        "description": "Print part of working in DERIT",
    },
    "DERIV": {
        "description": "Print part of working in DERIV",
    },
    "DERNVO": {
        "description": "Print part of working in DERNVO",
    },
    "DFORCE": {
        "description": "Force calculation specified, also print force matrix.",
    },
    "DFP": {
        "description": (
            "Use Davidson-Fletcher-Powell method to optimize " "geometries"
        ),
    },
    "DISEX": {
        "description": (
            "Distance for interactions in fine grid in COSMO. "
            "Defaults is 2.0, should converge results using "
            "smaller values if needed."
        ),
        "takes values": 1,
        "default": "2.0",
        "format": "{}={}",
    },
    "DISP": {
        "description": (
            "Print the hydrogen bonding and dispersion "
            "contributions to the heat of formation"
        ),
    },
    "DMAX": {
        "description": "Maximum stepsize in eigenvector following",
        "takes values": 1,
        "default": "2.0",
        "format": "{}={}",
    },
    "DOUBLET": {
        "description": "Doublet state required",
    },
    "DRC": {
        "description": (
            "Dynamic reaction coordinate calculation. The "
            "parameter, if given, is the cooling half-life in fs"
        ),
        "takes values": [0, 1],
        "default": "50",
        "format": "{}={}",
    },
    "DUMP": {
        "description": (
            "Write restart files every n seconds, or minutes or "
            'hours with "m" or "h" suffix'
        ),
        "takes values": 1,
        "default": "2h",
        "format": "{}={}",
    },
    "ECHO": {
        "description": "Data are echoed back before calculation starts",
    },
    "EF": {
        "description": ("Use the EigenFollowing routine for geometry " "optimization"),
    },
    "EIGEN": {
        "description": (
            "Print canonical eigenvectors instead of LMOs in a " "MOZYME calculations"
        ),
    },
    "EIGS": {
        "description": "Print all eigenvalues in ITER",
    },
    "ENPART": {"description": "Partition energy into components"},
    "EPS": {
        "description": "Dielectric constant in COSMO calculation",
        "takes values": 1,
        "default": "78.4",
        "format": "{}={}",
    },
    "ESP": {
        "description": "Electrostatic potential calculation",
    },
    "ESPRST": {
        "description": "Restart of electrostatic potential",
    },
    "ESR": {
        "description": "Calculate RHF spin density",
    },
    "EXCITED": {
        "description": "Optimize first excited singlet state",
    },
    "EXTERNAL": {
        "description": "Read parameters from a file",
        "takes values": 1,
        "default": '"file name"',
        "format": "{}={}",
    },
    "FIELD=(n.nn,m.mm,l.ll)": {
        "description": (
            "Apply an external electric field of a,b,c V/Å in the"
            "Cartesian directions"
        ),
        "takes values": 3,
        "default": "(0,0,1)",
        "format": "{}={}",
    },
    "FILL": {
        "description": ("In RHF open and closed shell, force M.O. n to be " "filled"),
        "takes values": 1,
        "default": "0",
        "format": "{}={}",
    },
    "FLEPO": {
        "description": "Print details of geometry optimization",
    },
    "FMAT": {
        "description": "Print details of working in FMAT",
    },
    "FOCK": {
        "description": "Print last Fock matrix",
    },
    "FREQCY": {
        "description": "Print symmetrized Hessian in a FORCE calculation",
    },
    "FORCE": {
        "description": "Calculate vibrational frequencies",
    },
    "FORCETS": {
        "description": (
            "Calculate the vibrational frequencies for atoms in a " "transition state"
        ),
    },
    "GEO-OK": {
        "description": "Override some safety checks",
    },
    "GEO_DAT": {
        "description": "Read in geometry from the file <text>",
        "takes values": 1,
        "default": '"file name"',
        "format": "{}={}",
    },
    "GEO_REF=<text>": {
        "description": "Read in a second geometry from the file <text>",
        "takes values": 1,
        "default": '"SELF"',
        "format": "{}={}",
    },
    "GNORM": {
        "description": (
            "Exit when the gradient norm drops below n.nn " "kcal/mol/Angstrom"
        ),
        "takes values": 1,
        "default": "10.0",
        "format": "{}={}",
    },
    "GRADIENTS": {
        "description": "Print all gradients",
    },
    "GRAPH": {
        "description": "Generate unformatted file for graphics",
    },
    "GRAPHF": {
        "description": (
            "Generate a formatted file for graphics suitable for " "Jmol and MOPETE."
        ),
    },
    "HCORE": {
        "description": (
            "Print all parameters used, the one-electron matrix, "
            "and two-electron integrals"
        ),
    },
    "HESSIAN": {
        "description": "Print Hessian from geometry optimization",
    },
    "HESS": {
        "description": "Options for calculating Hessian matrices in EF",
        "takes values": 1,
        "default": "1",
        "format": "{}={}",
        "allowed values": (0, 1, 2),
    },
    "H-PRIORITY": {
        "description": (
            "Heat of formation takes priority in DRC. Print "
            "whenever the energy changes by this many kcal/mol"
        ),
        "takes values": 1,
        "default": "0.1",
        "format": "{}={}",
    },
    "HTML": {
        "description": "Write a web-page for displaying and editing a protein",
    },
    "HYPERFINE": {
        "description": "Hyperfine coupling constants to be calculated",
    },
    "INT": {
        "description": "Make all coordinates internal coordinates",
    },
    "INVERT": {
        "description": "Reverse all optimization flags",
    },
    "IRC": {
        "description": "Intrinsic reaction coordinate calculation",
        "takes values": [0, 1],
        "default": "",
        "format": "{}={}",
    },
    "ISOTOPE": {
        "description": "Force matrix written to disk (channel 9 )",
    },
    "ITER": {
        "description": "Print details of working in ITER",
    },
    "ITRY": {
        "description": "Set limit of number of SCF iterations to n",
        "takes values": 1,
        "default": "2000",
        "format": "{}={}",
    },
    "IUPD": {
        "description": (
            "Mode of Hessian update in eigenvector following: "
            "0: skip, 1: Powell, 2: BFGS"
        ),
        "takes values": 1,
        "default": "2",
        "format": "{}={}",
        "allowed values": (0, 1, 2),
    },
    "KINETIC": {
        "description": "Excess kinetic energy added to DRC calculation",
        "takes values": 1,
        "default": 20,
        "format": "{}={}",
    },
    "KING": {
        "description": "Use Camp-King converger for SCF",
    },
    "LARGE": {
        "description": "Print expanded output",
    },
    "LBFGS": {
        "description": "Use the low-memory version of the BFGS optimizer",
    },
    "LET": {
        "description": "Override certain safety checks",
    },
    "LEWIS": {
        "description": "Print the Lewis structure",
    },
    "LINMIN": {
        "description": "Print details of line minimization",
    },
    "LOCALIZE": {
        "description": (
            "Print the localized orbitals. These are also called "
            "Natural Bond Orbitals or NBO"
        ),
    },
    "LOCATE-TS": {
        "description": (
            "Given reactants and products, locate the transition "
            "state connecting them"
        ),
    },
    "LOG": {
        "description": "Generate a log file",
    },
    "MECI": {
        "description": "Print details of MECI calculation",
    },
    "MERS=(n1,n2,n3)": {
        "description": "Keyword generated by MAKPOL for use with program BZ",
        "takes values": [1, 2, 3],
        "default": "(n1[,n2[,n3]])",
        "format": "{}={}",
    },
    "METAL": {
        "description": "Make specified atoms 100% ionic",
        "takes values": "0+",
        "default": "(Au(+3),Fe)",
        "format": "{}={}",
    },
    "MICROS": {
        "description": "Use specific microstates in the C.I.",
        "takes values": 1,
        "default": "1",
        "format": "{}={}",
    },
    "MINI": {
        "description": (
            "Reduce the size of the output by only printing " "specified atoms"
        ),
    },
    "MINMEP": {
        "description": "Minimize MEP minima in the plane defined",
    },
    "MMOK": {
        "description": "Use molecular mechanics correction to CONH bonds",
    },
    "MNDO": {
        "description": "Use the MNDO hamiltonian",
    },
    "MNDOD": {"description": "Use the MNDO-d hamiltonian"},
    "MODE": {
        "description": "In EF, follow Hessian mode no. n",
        "takes values": 1,
        "default": 1,
        "format": "{}={}",
    },
    "MOL_QMMM": {
        "description": ("Incorporate environmental effects in the QM/MM " "approach"),
    },
    "MOLDAT": {
        "description": "Print details of working in MOLDAT",
    },
    "MOLSYM": {
        "description": "Print details of working in MOLSYM",
    },
    "MOPAC": {
        "description": "Use old MOPAC definition for 2nd and 3rd atoms",
    },
    "MOZYME": {
        "description": (
            "Use the Localized Molecular Orbital method to speed " "up the SCF"
        ),
    },
    "MS": {
        "description": "In MECI, magnetic component of spin",
        "takes values": 1,
        "default": "1.5",
        "format": "{}={}",
        "allowed value test": "2*value",
    },
    "MULLIK": {
        "description": "Print the Mulliken population analysis",
    },
    "N**2": {
        "description": (
            "In excited state COSMO calculations, set the value " "of N**2"
        ),
        "takes values": 1,
        "default": 2,
        "format": "{}={}",
    },
    "NLLSQ": {
        "description": "Minimize gradients using NLLSQ",
    },
    "NOANCI": {
        "description": "Do not use analytical C.I. derivatives",
    },
    "NOCOMMENTS": {
        "description": (
            "Ignore all lines except ATOM, HETATM, and TER in PDB " "files"
        ),
    },
    "NOGPU": {
        "description": "Do not use GPU acceleration",
    },
    "NOLOG": {
        "description": "Suppress log file trail, where possible",
    },
    "NOMM": {
        "description": ("Do not use molecular mechanics correction to CONH " "bonds"),
    },
    "NONET": {
        "description": "NONET state required",
    },
    "NONR": {
        "description": "Do not use Newton-Raphson method in EF",
    },
    "NOOPT": {
        "description": ("Do not optimize the coordinates of all atoms " "(see OPT-X)"),
    },
    "NOOPT-X": {
        "description": ("Do not optimize the coordinates of all atoms of type " "X"),
    },
    "NOREOR": {
        "description": "In symmetry work, use supplied orientation",
    },
    "NORESEQ": {
        "description": (
            "Suppress the default re-sequencing of atoms to the " "PDB sequence"
        ),
    },
    "NOSWAP": {
        "description": "Do not allow atom swapping when GEO_REF is used",
    },
    "NOSYM": {
        "description": "Point-group symmetry set to C1",
    },
    "NOTER": {
        "description": 'Do not put "TER"s in PDB files',
    },
    "NOTHIEL": {
        "description": "Do not use Thiel's FSTMIN technique",
    },
    "NOTXT": {
        "description": "Remove any text from atom symbols",
    },
    "NOXYZ": {
        "description": "Do not print Cartesian coordinates",
    },
    "NSPA": {
        "description": (
            "Sets number of geometric segments in COSMO " "use 42, 92, 122 or 162"
        ),
        "takes values": 1,
        "default": 42,
        "format": "{}={}",
    },
    "NSURF": {
        "description": "Number of surfaces in an ESP calculation",
    },
    "OCTET": {
        "description": "Octet state required",
    },
    "OLDCAV": {
        "description": (
            "In COSMO, use the old Solvent Accessible Surface " "calculation"
        ),
    },
    "OLDENS": {
        "description": "Read initial density matrix off disk",
    },
    "OLDFPC": {
        "description": "Use the old fundamental physical constants",
    },
    "OLDGEO": {
        "description": "Previous geometry to be used",
    },
    "OMIN=n.nn": {
        "description": "In TS, minimum allowed overlap of eigenvectors",
        "takes values": 1,
        "default": 0.8,
        "format": "{}={}",
    },
    "OPEN": {
        "description": "Open-shell UHF or RHF calculation requested",
        "takes values": 2,
        "default": "(2,2)",
        "format": "{}{}",
    },
    "OPT": {
        "description": (
            "OPT: Optimize coordinates of all atoms -- or --"
            "OPT(text=n.nn): Optimize the coordinates of all atoms within"
            'n.nn Ångstroms of atoms labeled "text" -- or --'
            'OPT("Label1"[,"Label1"[,"Label1"...]]) where labels are residues'
        ),
        "takes values": [0, 1],
        "default": "",
        "format": "OPT({value})",
    },
    "OPT-X": {
        "description": "Optimize the coordinates of all atoms of type X",
        "takes values": 1,
        "default": "H",
        "format": "OPT-X={value}",
    },
    "OUTPUT": {
        "description": ("Reduce the amount of output (useful for large " "systems)"),
    },
    "P": {
        "description": (
            "An applied pressure of n.nn Newtons/m2 to be used"
            " A suffix of GPa may be used to indicate GPa=10^9 "
            "N/m^2"
        ),
        "takes values": 1,
        "default": "10.0GPa",
        "format": "{}={}",
    },
    "PDB": {
        "description": (
            "Input geometry is in protein data bank format. "
            "Arguments give nonstandard element name:atno, e.g. "
            "D:1 or LP:0"
        ),
        "takes values": 1,
        "default": "(D:1,LP:0)",
        "format": "{}{}",
    },
    "PDBOUT	": {
        "description": "Output geometry in pdb format",
    },
    "PECI": {
        "description": "C.I. involves paired excitations only",
    },
    "PI": {
        "description": "Resolve density matrix into σ, π, and δ components",
    },
    "pKa": {
        "description": (
            "Print the pKa for ionizable hydrogen atoms attached " "to oxygen atoms"
        ),
    },
    "PL": {
        "description": "Monitor convergence of density matrix in ITER",
    },
    "PM3": {
        "description": "Use the MNDO-PM3 Hamiltonian",
    },
    "PM6": {"description": "Use the PM6 Hamiltonian"},
    "PM6-D3": {
        "description": (
            "Use the PM6 Hamiltonian with Grimme's corrections " "for dispersion"
        ),
    },
    "PM6-DH+": {
        "description": (
            "Use the PM6 Hamiltonian with corrections for "
            "dispersion and hydrogen-bonding"
        ),
    },
    "PM6-DH2": {
        "description": (
            "Use the PM6 Hamiltonian with corrections for "
            "dispersion and hydrogen-bonding"
        ),
    },
    "PM6-DH2X": {
        "description": (
            "Use PM6 with corrections for dispersion and hydrogen "
            "and halogen bonding"
        ),
    },
    "PM6-D3H4": {
        "description": "Use PM6 with Řezáč and Hobza's D3H4 correction",
    },
    "PM6-D3H4X": {
        "description": ("Use PM6 with Brahmkshatriya, et al.'s D3H4X " "correction"),
    },
    "PMEP": {
        "description": "Complete semiempirical MEP calculation",
    },
    "PM7": {
        "description": "Use the PM7 Hamiltonian",
    },
    "PM7-TS": {
        "description": "Use the PM7-TS Hamiltonian (only for barrier heights)",
    },
    "PMEPR": {
        "description": "Complete semiempirical MEP in a plane to be defined",
    },
    "POINT": {
        "description": "Number of points in reaction path",
        "takes values": 1,
        "default": "40",
        "format": "{}={}",
    },
    "POINT1": {
        "description": ("Number of points in first direction in grid " "calculation"),
        "takes values": 1,
        "default": "11",
        "format": "{}={}",
    },
    "POINT2": {
        "description": ("Number of points in second direction in grid " "calculation"),
        "takes values": 1,
        "default": "11",
        "format": "{}={}",
    },
    "POLAR": {
        "description": ("Calculate first, second and third order " "polarizabilities"),
    },
    "POTWRT": {
        "description": "In ESP, write out electrostatic potential to unit 21",
    },
    "POWSQ": {
        "description": "Print details of working in POWSQ",
    },
    "PRECISE": {
        "description": "Criteria to be increased by 100 times",
    },
    "PRESSURE": {
        "description": "Apply pressure or tension to a solid or polymer",
    },
    "PRNT": {
        "description": (
            "Print details of geometry optimization in EF, 1 = " "most, 5=least"
        ),
        "takes values": 1,
        "default": 1,
        "format": "{}={}",
        "allowed values": (0, 1, 2, 3, 4, 5),
    },
    "PRTCHAR": {
        "description": "Print charges in ARC file",
    },
    "PRTINT": {
        "description": "Print interatomic distances",
    },
    "PRTMEP": {
        "description": "MEP contour data output to <filename>.mep",
    },
    "PRTXYZ": {
        "description": "Print Cartesian coordinates",
    },
    "PULAY": {
        "description": "Use Pulay's converger to obtain a SCF",
    },
    "QMMM": {
        "description": ("Incorporate environmental effects in the QM/MM " "approach"),
    },
    "QPMEP": {
        "description": "Charges derived from Wang-Ford type AM1 MEP",
    },
    "QUARTET": {
        "description": "Quartet state required",
    },
    "QUINTET": {
        "description": "Quintet state required",
    },
    "RAPID": {
        "description": (
            "In MOZYME geometry optimizations, only use atoms "
            "being optimized in the SCF"
        ),
    },
    "RECALC": {
        "description": "In EF, recalculate Hessian every n steps",
        "takes values": 1,
        "default": 10,
        "format": "{}={}",
    },
    "RE-LOCAL": {
        "description": (
            "During and at end of a MOZYME calculation, "
            "re-localize the LMOs. Default is just at end."
        ),
        "takes values": [0, 1],
        "default": 1,
        "format": "{}={}",
    },
    "RELSCF": {
        "description": "Default SCF criterion multiplied by n",
    },
    "REORTHOG": {
        "description": (
            "In MOZYME, re-orthogonalize LMO's each 10 SCF " "calculations."
        ),
    },
    "RESEQ": {
        "description": "Re-arrange the atoms to match the PDB convention",
    },
    "RESIDUES": {
        "description": (
            "Label each atom in a polypeptide with the amino acid " "residue."
        ),
    },
    "RESTART": {
        "description": "Calculation restarted",
    },
    "RHF": {
        "description": "Use Restricted Hartree-Fock methods",
    },
    "RM1": {
        "description": "Use the RM1 Hamiltonian",
    },
    "RMAX": {
        "description": "In TS, maximum allowed ratio for energy change",
        "takes values": 1,
        "default": 4.0,
        "format": "{}={}",
    },
    "RMIN": {
        "description": "In TS, minimum allowed ratio for energy change",
        "takes values": 1,
        "default": 0.0,
        "format": "{}={}",
    },
    "ROOT": {
        "description": (
            "Root n to be optimized in a C.I. calculation. "
            "Either an integer or with a symmetry label, e.g."
            "2T2g"
        ),
        "takes values": 1,
        "default": "1",
        "format": "{}={}",
    },
    "RSCAL": {
        "description": "In EF, scale p-RFO to trust radius",
    },
    "RSOLV=n.nn": {
        "description": "Effective radius of solvent in COSMO",
    },
    "SADDLE": {
        "description": "Optimize transition state",
    },
    "SCALE": {
        "description": "Scaling factor for van der waals distance in ESP",
    },
    "SCFCRT=n.nn": {
        "description": "Default SCF criterion replaced by the value supplied",
    },
    "SCINCR=n.nn": {
        "description": "Increment between layers in ESP",
    },
    "SEPTET": {
        "description": "Septet state required",
    },
    "SETPI": {
        "description": ("In MOZYME, some π bonds are explicitly set by the " "user"),
    },
    "SETUP": {
        "description": "Extra keywords to be read from setup file",
    },
    "SEXTET": {
        "description": "Sextet state required",
    },
    "SHIFT=n.nn": {
        "description": "a damping factor of n defined to start SCF",
    },
    # 'SHUT <file>': {
    #     'description': ('Send a command to MOPAC to make a restart and '
    #                     'density file, then stop.'),
    # },
    "SIGMA": {
        "description": "Minimize gradients using SIGMA",
    },
    "SINGLET": {
        "description": "Singlet state required",
    },
    "SITE=(text)": {
        "description": "Define ionization state of residues in proteins",
    },
    "SLOG=n.nn": {
        "description": ("In L-BFGS optimization, use fixed step of length " "n .nn"),
    },
    "SLOPE": {
        "description": "Multiplier used to scale MNDO charges",
    },
    "SMOOTH": {
        "description": (
            "In a GRID calculation, remove artifacts caused by "
            "the order in which points are calculated"
        ),
    },
    "SNAP": {
        "description": "Increase precision of symmetry angles",
    },
    "SPARKLE": {
        "description": "Use sparkles instead of atoms with basis sets",
    },
    "SPIN": {
        "description": "Print final UHF spin matrix",
    },
    "START_RES(text)": {
        "description": (
            "Define starting residue numbers in a protein, if "
            "different from the default"
        ),
    },
    "STATIC": {
        "description": "Calculate Polarizability using electric fields",
    },
    "STEP": {
        "description": "Step size in path",
    },
    "STEP1=n.nnn": {
        "description": "Step size n for first coordinate in grid calculation",
    },
    "STEP2=n.nnn": {
        "description": "Step size n for second coordinate in grid calculation",
    },
    "STO3G": {
        "description": "Deorthogonalize orbitals in STO-3G basis",
    },
    "SUPER": {
        "description": "Print superdelocalizabilities",
    },
    "SYBYL": {
        "description": "Output a file for use by Tripos's SYBYL program",
    },
    "SYMAVG": {
        "description": "Average symmetry equivalent ESP charges",
    },
    "SYMOIR": {
        "description": (
            "Print characters of eigenvectors and print number of " "I.R.s"
        ),
    },
    "SYMTRZ": {
        "description": "Print details of working in subroutine SYMTRZ.",
    },
    "SYMMETRY": {
        "description": "Impose symmetry conditions",
    },
    "T=n[M,H,D]": {
        "description": "A time of n seconds requested",
    },
    "THERMO(nnn,mmm,lll)": {
        "description": "Perform a thermodynamics calculation",
    },
    "THREADS=n": {
        "description": (
            "Set the number of threads to be used in " "parallelization to n"
        ),
    },
    "TIMES": {
        "description": "Print times of various stages",
    },
    "T-PRIORITY=n.nn": {
        "description": "Time takes priority in DRC",
    },
    "TRANS=n": {
        "description": (
            "The system is a transition state (used in " "thermodynamics calculation)"
        ),
    },
    "TRIPLET": {
        "description": "Triplet state required",
    },
    "TS": {
        "description": "Using EF routine for TS search",
    },
    "UHF": {
        "description": "Use the Unrestricted Hartree-Fock method",
    },
    "VDW(text)": {
        "description": ("Van der waals radius for atoms in COSMO defined by " "user"),
    },
    "VDWM(text)": {
        "description": ("Van der waals radius for atoms in MOZYME defined by " "user"),
    },
    "VECTORS": {
        "description": "Print final eigenvectors",
    },
    "VELOCITY": {
        "description": ("Supply the initial velocity vector in a DRC " "calculation"),
    },
    "WILLIAMS": {
        "description": "Use Williams surface",
    },
    "X-PRIORITY=n.nn": {
        "description": "Geometry changes take priority in DRC",
    },
    "XENO": {
        "description": ("Allow non-standard residues in proteins to be " "labeled."),
    },
    "XYZ": {
        "description": "Do all geometric operations in Cartesian coordinates",
    },
    "Z=n": {
        "description": "Number of mers in a cluster",
    },
}

properties = {
    "AO_ATOMINDEX": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "atom for AO",
        "dimensionality": ["n_aos"],
        "type": "integer",
    },
    "AO_ZETA": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "Slater exponent",
        "dimensionality": ["n_aos"],
        "type": "float",
    },
    "AREA": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "surface area",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å^2",
    },
    "ATOM_CHARGES": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "atom charges",
        "dimensionality": ["n_atoms"],
        "type": "float",
    },
    "ATOM_CORE": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "number of valence electrons",
        "dimensionality": ["n_atoms"],
        "type": "integer",
    },
    "ATOM_EL": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "atomic symbol",
        "dimensionality": ["n_atoms"],
        "type": "string",
    },
    "ATOM_PQN": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "principal quantum number",
        "dimensionality": ["n_aos"],
        "type": "integer",
    },
    "ATOM_SYMTYPE": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "atomic orbital angular shape",
        "dimensionality": ["n_aos"],
        "type": "string",
    },
    "ATOM_X": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "x, y, z coordinates",
        "dimensionality": [3, "n_atoms"],
        "type": "float",
        "units": "Å",
    },
    "ATOM_X_FORCE": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "reoriented coordinates",
        "dimensionality": [3, "n_atoms"],
        "type": "float",
        "units": "Å",
    },
    "ATOM_X_OPT": {
        "calculation": ["optimization", "thermodynamics", "vibrations"],
        "description": "optimized x, y, z coordinates",
        "dimensionality": [3, "n_atoms"],
        "type": "float",
        "units": "Å",
    },
    "ATOM_X_UPDATED": {
        "calculation": ["optimization"],
        "description": "trajectory coordinates",
        "dimensionality": ["nsteps", [3, "n_atoms"]],
        "type": "float",
        "units": "Å",
    },
    "COMMENTS": {
        "calculation": [],
        "description": "User comment line",
        "dimensionality": "scalar",
        "type": "string",
    },
    "CPU_TIME": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "cpu time for calculation",
        "dimensionality": "scalar",
        "type": "float",
        "units": "s",
    },
    "DATE": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "date and time of calculation",
        "dimensionality": "scalar",
        "type": "date_time",
    },
    "DENSITY": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "density",
        "dimensionality": "scalar",
        "type": "float",
        "units": "g/mL",
    },
    "DIPOLE": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "dipole moment",
        "dimensionality": "scalar",
        "type": "float",
        "units": "debye",
    },
    "DIP_VEC": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "dipole vector",
        "dimensionality": [3],
        "type": "float",
        "units": "debye",
    },
    "EIGENVALUES": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital energies",
        "dimensionality": ["n_aos"],
        "type": "float",
        "units": "eV",
    },
    "ALPHA_EIGENVALUES": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital energies",
        "dimensionality": ["n_aos"],
        "type": "float",
        "units": "eV",
    },
    "BETA_EIGENVALUES": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital energies",
        "dimensionality": ["n_aos"],
        "type": "float",
        "units": "eV",
    },
    "LMO_ENERGY_LEVELS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "Localized MO energies",
        "dimensionality": ["n_aos"],
        "type": "float",
        "units": "eV",
    },
    "EIGENVECTORS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital coefficients",
        "dimensionality": ["n_aos"],
        "type": "float",
    },
    "ALPHA_EIGENVECTORS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital coefficients",
        "dimensionality": ["n_aos"],
        "type": "float",
    },
    "BETA_EIGENVECTORS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital coefficients",
        "dimensionality": ["n_aos"],
        "type": "float",
    },
    "LMO_VECTORS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "localized molecular orbital coefficients",
        "dimensionality": ["n_aos"],
        "type": "float",
    },
    "EMPIRICAL_FORMULA": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "empirical formula",
        "dimensionality": "scalar",
        "type": "string",
    },
    "ENERGY_ELECTRONIC": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "electronic energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "eV",
    },
    "ENERGY_NUCLEAR": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "nuclear repulsion energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "eV",
    },
    "DIEL_ENER": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "the dielectric energy from COSMO",
        "dimensionality": "scalar",
        "type": "float",
        "units": "eV",
    },
    "ENTHALPY_TOT": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "enthalpy",
        "dimensionality": [
            "n_temps",
        ],
        "type": "float",
        "units": "cal/mol",
    },
    "ENTROPY_TOT": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "entropy",
        "dimensionality": [
            "n_temps",
        ],
        "type": "float",
        "units": "cal/K/mol",
    },
    "GRADIENTS": {
        "calculation": ["optimization", "thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": [3, "n_atoms"],
        "type": "float",
        "units": "kcal/mol/Å",
    },
    "GRADIENT_NORM": {
        "calculation": ["optimization", "thermodynamics", "vibrations"],
        "description": "final norm of the gradient",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol/Å",
    },
    "GRADIENT_UPDATED": {
        "calculation": ["optimization"],
        "description": "forces in trajectory",
        "dimensionality": ["nsteps", [3, "n_atoms"]],
        "type": "float",
        "units": "kcal/mol/Å",
    },
    "HEAT_CAPACITY_TOT": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "heat capacity",
        "dimensionality": [
            "n_temps",
        ],
        "type": "float",
        "units": "cal/K/mol",
    },
    "HEAT_OF_FORMATION": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "heat of formation",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol",
    },
    "HEAT_OF_FORM_UPDATED": {
        "calculation": ["optimization"],
        "description": "heat of formation in trajectory",
        "dimensionality": ["nsteps"],
        "type": "float",
        "units": "kcal/mol",
    },
    "HESSIAN_MATRIX": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_dof", "n_dof"],
        "type": "float",
        "units": "mdynes/Å/Da",
    },
    "H_O_F(T)": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_temps"],
        "type": "float",
        "units": "kcal/mol",
    },
    "INT_FORCE_CONSTS": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "force constants for internals",
        "dimensionality": ["n_dofs"],
        "type": "float",
        "units": "mdyne/Å",
    },
    "INITIAL_TRANS_VECTS": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "initial translation vectors",
        "dimensionality": [9],
        "type": "float",
        "units": "Å",
    },
    "IONIZATION_POTENTIAL": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "ionization potential (IP)",
        "dimensionality": "scalar",
        "type": "float",
        "units": "eV",
    },
    "ISOTOPIC_MASSES": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_atoms"],
        "type": "float",
        "units": "Da",
    },
    "KEYWORDS": {
        "calculation": [],
        "description": "",
        "dimensionality": "scalar",
        "type": "string",
    },
    "M.O.SYMMETRY_LABELS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "molecular orbital symmetries",
        "dimensionality": ["n_mos"],
        "type": "string",
    },
    "ALPHA_M.O.SYMMETRY_LABELS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "molecular orbital symmetries",
        "dimensionality": ["n_mos"],
        "type": "string",
    },
    "BETA_M.O.SYMMETRY_LABELS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "molecular orbital symmetries",
        "dimensionality": ["n_mos"],
        "type": "string",
    },
    "METHOD": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "hamiltonian",
        "dimensionality": "scalar",
        "type": "string",
    },
    "MOLECULAR_ORBITAL_OCCUPANCIES": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital occupancies",
        "dimensionality": ["n_mos"],
        "type": "float",
    },
    "ALPHA_MOLECULAR_ORBITAL_OCCUPANCIES": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital occupancies",
        "dimensionality": ["n_mos"],
        "type": "float",
    },
    "BETA_MOLECULAR_ORBITAL_OCCUPANCIES": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "orbital occupancies",
        "dimensionality": ["n_mos"],
        "type": "float",
    },
    "MOLECULAR_WEIGHT": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "molecular weight",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Da",
    },
    "MOPAC_VERSION": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "MOPAC version",
        "dimensionality": "scalar",
        "type": "string",
    },
    "NORMAL_MODES": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_dof", "n_dof"],
        "type": "float",
    },
    "NORMAL_MODE_SYMMETRY_LABELS": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_dof"],
        "type": "string",
    },
    "NUMBER_SCF_CYCLES": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "number of scf's",
        "dimensionality": "scalar",
        "type": "integer",
    },
    "NUM_ALPHA_ELECTRONS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "number of spin-up electrons",
        "dimensionality": "scalar",
        "type": "integer",
    },
    "NUM_BETA_ELECTRONS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "number of spin-down electrons",
        "dimensionality": "scalar",
        "type": "integer",
    },
    "NUM_ELECTRONS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "number of electrons",
        "dimensionality": "scalar",
        "type": "integer",
    },
    "ORIENTATION_ATOM_X": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": [3, "n_atoms"],
        "type": "float",
        "units": "Å",
    },
    "OVERLAP_MATRIX": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "the AO overlap matrix",
        "dimensionality": ["triangular", "n_aos", "n_aos"],
        "type": "float",
    },
    "DENSITY_MATRIX": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "the density matrix",
        "dimensionality": ["triangular", "n_aos", "n_aos"],
        "type": "float",
    },
    "POINT_GROUP": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "the molecular symmetry",
        "dimensionality": "scalar",
        "type": "string",
    },
    "PRI_MOM_OF_I": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": [3],
        "type": "float",
        "units": "10^-40*g*cm^2",
    },
    "RESTRAINING_PRESSURE": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "the pressure in the three lattice directions",
        "dimensionality": [3],
        "type": "float",
        "units": "GPa",
    },
    "ROTAT_CONSTS": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": [3],
        "type": "float",
        "units": "1/cm",
    },
    "SET_OF_MOS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "set of MOs",
        "dimensionality": [2],
        "type": "string",
    },
    "SET_OF_ALPHA_MOS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "set of MOs",
        "dimensionality": [2],
        "type": "string",
    },
    "SET_OF_BETA_MOS": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "set of MOs",
        "dimensionality": [2],
        "type": "string",
    },
    "SIZE_OF_ACTIVE_SPACE": {
        "calculation": [
            "optimization",
        ],
        "description": "size of active space",
        "dimensionality": [1],
        "type": "string",
    },
    "NUMBER_MICROSTATES": {
        "calculation": [
            "optimization",
        ],
        "description": "number of microstates",
        "dimensionality": [1],
        "type": "string",
    },
    "MICROSTATE_CONFIGURATIONS": {
        "calculation": [
            "optimization",
        ],
        "description": "microstate configurations",
        "dimensionality": [
            "n_microstates",
        ],
        "type": "integer",
    },
    "STATE_REQUESTED": {
        "calculation": [
            "optimization",
        ],
        "description": "state requested",
        "dimensionality": [1],
        "type": "integer",
    },
    "STATE_DEGENERACY": {
        "calculation": [
            "optimization",
        ],
        "description": "number of microstates",
        "dimensionality": [1],
        "type": "string",
    },
    "STATE_VECTOR": {
        "calculation": [
            "optimization",
        ],
        "description": "state vector",
        "dimensionality": [1],
        "type": "float",
    },
    "STATE": {
        "calculation": [
            "optimization",
        ],
        "description": "state",
        "dimensionality": [1],
        "type": "string",
    },
    "STATE_ENERGY_ABSOLUTE": {
        "calculation": [
            "optimization",
        ],
        "description": "state energy absolute",
        "dimensionality": [1],
        "type": "string",
    },
    "STATE_ENERGY_RELATIVE": {
        "calculation": [
            "optimization",
        ],
        "description": "state",
        "dimensionality": [1],
        "type": "string",
    },
    "SPIN_COMPONENT": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "spin component",
        "dimensionality": [3],
        "type": "float",
    },
    "THERMODYNAMIC_PROPERTIES_TEMPS": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_temps"],
        "type": "float",
        "units": "K",
    },
    "TITLE": {
        "calculation": [],
        "description": "title of calculation",
        "dimensionality": "scalar",
        "type": "string",
    },
    "TOTAL_DENSITY_MATRIX": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "density matrix",
        "dimensionality": ["n_mos", "n_mos"],
        "shape": "triangular",
        "type": "float",
    },
    "ALPHA_DENSITY_MATRIX": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "density matrix",
        "dimensionality": ["n_mos", "n_mos"],
        "shape": "triangular",
        "type": "float",
    },
    "BETA_DENSITY_MATRIX": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "density matrix",
        "dimensionality": ["n_mos", "n_mos"],
        "shape": "triangular",
        "type": "float",
    },
    "TOTAL_ENERGY": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "total energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "eV",
    },
    "TOTAL_SPIN": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "spin",
        "dimensionality": "scalar",
        "type": "float",
    },
    "TRANS_VECTS": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "initial translation vectors",
        "dimensionality": [9],
        "type": "float",
        "units": "Å",
    },
    "TRANS_VECTS_UPDATED": {
        "calculation": [
            "energy",
            "optimization",
        ],
        "description": "initial translation vectors",
        "dimensionality": [9],
        "type": "float",
        "units": "Å",
    },
    "VIB._EFF_MASS": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_dof"],
        "type": "float",
        "units": "Da",
    },
    "VIB._FREQ": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_dof"],
        "type": "float",
        "units": "1/cm",
    },
    "VIB._RED_MASS": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_dof"],
        "type": "float",
        "units": "Da",
    },
    "VIB._TRAVEL": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_dof"],
        "type": "float",
        "units": "Å",
    },
    "VIB._T_DIP": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "",
        "dimensionality": ["n_dof"],
        "type": "float",
        "units": "electrons",
    },
    "VOLUME": {
        "calculation": ["energy", "optimization", "thermodynamics", "vibrations"],
        "description": "volume",
        "dimensionality": "scalar",
        "type": "float",
        "units": "Å^3",
    },
    "ZERO_POINT_ENERGY": {
        "calculation": ["thermodynamics", "vibrations"],
        "description": "zero point energy",
        "dimensionality": "scalar",
        "type": "float",
        "units": "kcal/mol",
    },
}



# -*- coding: utf-8 -*-
"""Control parameters for a MOPAC thermodynamics calculation
"""

import logging
import mopac_step

logger = logging.getLogger(__name__)


class ThermodynamicsParameters(mopac_step.EnergyParameters):
    """The control parameters for creating a structure from SMILES"""

    parameters = {
        "Tmin": {
            "default": "200",
            "kind": "float",
            "default_units": "K",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Minimum temperature:",
            "help_text": (
                "The minimum temperature for the thermodynamic " "functions."
            ),
        },
        "Tmax": {
            "default": "400",
            "kind": "float",
            "default_units": "K",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Maximum temperature:",
            "help_text": (
                "The maximum temperature for the thermodynamic " "functions."
            ),
        },
        "Tstep": {
            "default": "10",
            "kind": "float",
            "default_units": "K",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Temperature interval:",
            "help_text": (
                "The interval between temperatures for the " "thermodynamic functions."
            ),
        },
        "trans": {
            "default": "0",
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "d",
            "description": "Number of internal rotations to ignore:",
            "help_text": (
                "The number of internal rotations to ignore. "
                "A corresponding number of the lowest modes "
                "will be ignored, which is a first approximation."
            ),
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={
                **ThermodynamicsParameters.parameters,
                **mopac_step.structure_handling_parameters,
                **defaults,
            },
            data=data,
        )



# -*- coding: utf-8 -*-

"""Setup and run MOPAC"""

import copy
import csv
import logging
from pathlib import Path
import textwrap

from tabulate import tabulate

import mopac_step
import seamm
import seamm.data
from seamm_util import Q_, units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("mopac")


class Energy(seamm.Node):
    def __init__(self, flowchart=None, title="Energy", extension=None):
        """Initialize the node"""

        logger.debug("Creating Energy {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self.parameters = mopac_step.EnergyParameters()

        self.description = "A single point energy calculation"

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return mopac_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return mopac_step.__git_revision__

    def description_text(self, P=None):
        """Prepare information about what this node will do"""

        if not P:
            P = self.parameters.values_to_dict()

        text = "Calculated with {hamiltonian}, converged to "
        # Convergence
        if P["convergence"] == "normal":
            text += "the 'normal' level of 1.0e-04 kcal/mol."
        elif P["convergence"] == "precise":
            text += "the 'precise' level of 1.0e-06 kcal/mol."
        elif P["convergence"] == "relative":
            text += "a factor of {relative} times the normal criterion."
        elif P["convergence"] == "absolute":
            text += "converged to {absolute}."

        if self.parameters["uhf"].is_expr:
            text += (
                " Whether to use spin-unrestricted SCF (UHF) for closed-shell molecules"
                "will be determined by '{uhf}'."
            )
        elif self.parameters["uhf"].get():
            text += " The SCF will be spin-unrestricted (UHF) for all molecules."
        else:
            text += (
                " The SCF will be restricted for closed-shell molecules (RHF) and "
                "spin-unrestricted (UHF) for all others."
            )

        # MOZYME localized molecular orbitals.
        if P["MOZYME"] == "always":
            text += (
                "\n\nThe SCF will be solved using localized molecular orbitals "
                "(MOZYME), which is faster than the traditional method for larger "
                "systems."
            )
            used_mozyme = True
        elif P["MOZYME"] == "for larger systems":
            text += (
                "\n\nThe SCF will be solved using localized molecular orbitals "
                "(MOZYME) for systems with {nMOZYME} atoms or more. This method is "
                "faster than the traditional method for larger systems."
            )
            used_mozyme = True
        else:
            used_mozyme = False

        if used_mozyme:
            follow_up = P["MOZYME follow-up"]
            if "exact" in follow_up:
                text += (
                    " The energy given by MOZYME slowly accumulates error due to the "
                    "increasing non-orthogonality of the localized orbitals after "
                    "many iterations. A single point energy using the traditional "
                    "method will be run to get the correct energy."
                )
            elif "new" in follow_up:
                text += (
                    " The energy given by MOZYME slowly accumulates error due to the "
                    "increasing non-orthogonality of the localized orbitals after "
                    "many iterations. A single point energy using fresh localized "
                    "orbitals will be run to get the correct energy."
                )
            elif follow_up == "none":
                text += (
                    " The energy given by MOZYME slowly accumulates error due to the "
                    "increasing non-orthogonality of the localized orbitals after "
                    "many iterations. No follow-up calculation will be done, so be "
                    "careful with the final energies produced."
                )
                used_mozyme = False
            else:
                logger.error(f"Don't recognize the MOZYME follow-up: '{follow_up}'")

        # Handle COSMO
        if self.parameters["COSMO"].is_expr:
            text += (
                "\n\n'{COSMO}' will determine whether to use the COSMO solvation "
                "model. If it is used the parameters will be "
            )
        elif self.parameters["COSMO"].get():
            text += "\n\nThe COSMO solvation model will be used with "

        if self.parameters["COSMO"].is_expr or self.parameters["COSMO"].get():
            text += (
                "dielectric constant = {eps}, solvent radius = {rsolve}, "
                "{nspa} grid points per atom, and a cutoff of {disex}."
            )

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self):
        """Get the input for an energy calculation for MOPAC"""
        system, configuration = self.get_system_configuration(None)
        references = self.parent.references

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        # Save the description for later printing
        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=self.indent))

        # Start gathering the keywords
        keywords = copy.deepcopy(P["extra keywords"])
        keywords.append("1SCF")
        keywords.append(P["hamiltonian"])

        if P["hamiltonian"] == "AM1":
            elements = configuration.atoms.symbols
            references.cite(
                raw=self.parent._bibliography["Dewar_1985c"],
                alias="Dewar_1985c",
                module="mopac_step",
                level=1,
                note="Main reference for AM1 + C, H, N, O.",
            )
            for element in ("F", "Cl", "Br", "I"):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Dewar_1988"],
                        alias="Dewar_1988",
                        module="mopac_step",
                        level=1,
                        note="AM1 parameters for F, Cl, Br, I.",
                    )
                    break
            if "Al" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1990"],
                    alias="Dewar_1990",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Al.",
                )
            if "Si" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1987b"],
                    alias="Dewar_1987b",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Si.",
                )
            if "P" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1989"],
                    alias="Dewar_1989",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for P.",
                )
            if "S" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1990b"],
                    alias="Dewar_1990b",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for S.",
                )
            if "Zn" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1988b"],
                    alias="Dewar_1988b",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Zn.",
                )
            if "Ge" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1989b"],
                    alias="Dewar_1989b",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Ge.",
                )
            if "Mo" in elements:
                references.cite(
                    raw=self.parent._bibliography["Voityuk_2000"],
                    alias="Voityuk_2000",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Mo.",
                )
            if "Hg" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1989c"],
                    alias="Dewar_1989c",
                    module="mopac_step",
                    level=1,
                    note="AM1 parameters for Hg.",
                )
            for element in (
                "Li",
                "Be",
                "Na",
                "Mg",
                "K",
                "Ca",
                "Ga",
                "As",
                "Se",
                "Rb",
                "Sr",
                "In",
                "Sn",
                "Sb",
                "Te",
                "Cs",
                "Ba",
                "Pb",
                "Bi",
            ):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Stewart_2004"],
                        alias="Stewart_2004",
                        module="mopac_step",
                        level=1,
                        note="AM1 parameterization for main-group elements.",
                    )
                    break
        elif P["hamiltonian"] == "MNDO" or P["hamiltonian"] == "MNDOD":
            elements = configuration.atoms.symbols
            references.cite(
                raw=self.parent._bibliography["Dewar_1977"],
                alias="Dewar_1977",
                module="mopac_step",
                level=1,
                note="Main reference for MNDO + C, H, N, O.",
            )
            if "Be" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1978"],
                    alias="Dewar_1978",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Be.",
                )
            if "B" in elements or "Al" in elements:
                if "B" in elements or P["hamiltonian"] == "MNDO":
                    references.cite(
                        raw=self.parent._bibliography["Davis_1981"],
                        alias="Davis_1981",
                        module="mopac_step",
                        level=1,
                        note="MNDO parameters for B and Al.",
                    )
            if "F" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1978b"],
                    alias="Dewar_1978b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for F.",
                )
            if "Si" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1986"],
                    alias="Dewar_1986",
                    module="mopac_step",
                    level=1,
                    note="Revised MNDO parameters for Si.",
                )
            if "P" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1978b"],
                    alias="Dewar_1978b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for P.",
                )
            if "S" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1986b"],
                    alias="Dewar_1986b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for S.",
                )
            if "Cl" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1983"],
                    alias="Dewar_1983",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Cl.",
                )
            if "Zn" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1986c"],
                    alias="Dewar_1986c",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Zn.",
                )
            if "Ge" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1987"],
                    alias="Dewar_1987",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Ge.",
                )
            if "Br" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1983b"],
                    alias="Dewar_1983b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Br.",
                )
            if "Sn" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1984"],
                    alias="Dewar_1984",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Sn.",
                )
            if "I" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1984b"],
                    alias="Dewar_1984b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for I.",
                )
            if "Hg" in elements and P["hamiltonian"] == "MNDO":
                references.cite(
                    raw=self.parent._bibliography["Dewar_1985"],
                    alias="Dewar_1985",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Hg.",
                )
            if "Pb" in elements:
                references.cite(
                    raw=self.parent._bibliography["Dewar_1985b"],
                    alias="Dewar_1985b",
                    module="mopac_step",
                    level=1,
                    note="MNDO parameters for Pb.",
                )
            for element in (
                "Na",
                "Mg",
                "K",
                "Ca",
                "Ga",
                "As",
                "Se",
                "Rb",
                "Sr",
                "In",
                "Sb",
                "Te",
                "Cs",
                "Ba",
                "Tl",
                "Bi",
            ):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Stewart_2004"],
                        alias="Stewart_2004",
                        module="mopac_step",
                        level=1,
                        note="MNDO parameterization for main-group elements.",
                    )
                    break
            if P["hamiltonian"] == "MNDOD":
                for element in (
                    "Al",
                    "Si",
                    "P",
                    "S",
                    "Cl",
                    "Br",
                    "I",
                    "Zn",
                    "Cd",
                    "Hg",
                ):
                    if element in elements:
                        references.cite(
                            raw=self.parent._bibliography["Thiel_1992"],
                            alias="Thiel_1992",
                            module="mopac_step",
                            level=1,
                            note=("MNDO-D formalism for d-orbitals."),
                        )
                        references.cite(
                            raw=self.parent._bibliography["Thiel_1996"],
                            alias="Thiel_1996",
                            module="mopac_step",
                            level=1,
                            note=(
                                "MNDO-D, parameters for Al, Si, P, S, Cl, Br, "
                                "I, Zn, Cd, and Hg."
                            ),
                        )
                        break
        elif P["hamiltonian"] == "PM3":
            elements = configuration.atoms.symbols
            references.cite(
                raw=self.parent._bibliography["Stewart_1989"],
                alias="Stewart_1989",
                module="mopac_step",
                level=1,
                note="The citation for the MOPAC parameterization.",
            )
            for element in (
                "Be",
                "Mg",
                "Zn",
                "Ga",
                "Ge",
                "As",
                "Se",
                "Cd",
                "In",
                "Sn",
                "Sb",
                "Te",
                "Hg",
                "Tl",
                "Pb",
                "Bi",
            ):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Stewart_1991"],
                        alias="Stewart_1991",
                        module="mopac_step",
                        level=1,
                        note="The citation for the MOPAC parameterization.",
                    )
                    break
            if "Li" in elements:
                references.cite(
                    raw=self.parent._bibliography["Anders_1993"],
                    alias="Anders_1993",
                    module="mopac_step",
                    level=1,
                    note="The citation for the MOPAC parameterization.",
                )
            for element in ("B", "Na", "K", "Ca", "Rb", "Sr", "Cs", "Ba"):
                if element in elements:
                    references.cite(
                        raw=self.parent._bibliography["Stewart_2004"],
                        alias="Stewart_2004",
                        module="mopac_step",
                        level=1,
                        note="The citation for the MOPAC parameterization.",
                    )
                    break
        elif "PM6" in P["hamiltonian"]:
            references.cite(
                raw=self.parent._bibliography["Stewart_2007"],
                alias="Stewart_2007",
                module="mopac_step",
                level=1,
                note="The PM6 parameterization in MOPAC.",
            )
            if P["hamiltonian"] == "PM6-D3":
                references.cite(
                    raw=self.parent._bibliography["Grimme_2010"],
                    alias="Grimme_2010",
                    module="mopac_step",
                    level=1,
                    note="Dispersion correction by Grimme, et al.",
                )
            if P["hamiltonian"] == "PM6-DH+":
                references.cite(
                    raw=self.parent._bibliography["Korth_2010"],
                    alias="Korth_2010",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-bonding correction by Korth.",
                )
            if "PM6-DH2" in P["hamiltonian"]:
                references.cite(
                    raw=self.parent._bibliography["Korth_2009"],
                    alias="Korth_2009",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-bonding and dispersion correction.",
                )
                references.cite(
                    raw=self.parent._bibliography["Rezac_2009"],
                    alias="Rezac_2009",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-bonding and dispersion correction.",
                )
            if P["hamiltonian"] == "PM6-DH2x":
                references.cite(
                    raw=self.parent._bibliography["Rezac_2011"],
                    alias="Rezac_2011",
                    module="mopac_step",
                    level=1,
                    note="Halogen-bonding correction.",
                )
            if "PM6-D3H4" in P["hamiltonian"]:
                references.cite(
                    raw=self.parent._bibliography["Rezac_2011"],
                    alias="Rezac_2011",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-bonding and dispersion correction.",
                )
                references.cite(
                    raw=self.parent._bibliography["Vorlova_2015"],
                    alias="Vorlova_2015",
                    module="mopac_step",
                    level=1,
                    note="Hydrogen-hydrogen repulsion correction.",
                )
            if P["hamiltonian"] == "PM6-D3H4x":
                references.cite(
                    raw=self.parent._bibliography["Brahmkshatriya_2013"],
                    alias="Brahmkshatriya_2013",
                    module="mopac_step",
                    level=1,
                    note="Halogen-oxygen and halogen-nitrogen correction.",
                )
        elif "PM7" in P["hamiltonian"]:
            references.cite(
                raw=self.parent._bibliography["Stewart_2012"],
                alias="Stewart_2012",
                module="mopac_step",
                level=1,
                note="The PM7 parameterization in MOPAC.",
            )
        elif P["hamiltonian"] == "RM1":
            references.cite(
                raw=self.parent._bibliography["Rocha_2006"],
                alias="Rocha_2006",
                module="mopac_step",
                level=1,
                note="RM1 parameterization.",
            )

        # which structure? may need to set default first...
        if P["structure"] == "default":
            if self._id[-1] == "1":
                structure = "initial"
            else:
                structure = "current"
        elif self._id[-1] == "1":
            structure = "initial"
        elif P["structure"] == "current":
            structure = "current"

        if structure == "current":
            keywords.append("OLDGEO")

        if P["convergence"] == "normal":
            pass
        elif P["convergence"] == "precise":
            keywords.append("PRECISE")
        elif P["convergence"] == "relative":
            keywords.append("RELSCF=" + P["relative"])
        elif P["convergence"] == "absolute":
            keywords.append("SCFSCRT=" + P["absolute"])
        else:
            raise RuntimeError(
                "Don't recognize convergence '{}'".format(P["convergence"])
            )

        if P["uhf"]:
            keywords.append("UHF")

        if P["MOZYME"] == "always":
            keywords.append("MOZYME")
        elif (
            P["MOZYME"] == "for larger systems"
            and configuration.n_atoms >= P["nMOZYME"]
        ):
            keywords.append("MOZYME")

        if P["COSMO"]:
            keywords.append(f"EPS={P['eps']}")
            rsolve = P["rsolve"].to("Å").magnitude
            keywords.append(f"RSOLVE={rsolve}")
            keywords.append(f"NSPA={P['nspa']}")
            keywords.append(f"DISEX={P['disex']}")

        if P["calculate gradients"]:
            keywords.append("GRADIENTS")

        # Add any extra keywords so that they appear at the end
        metadata = mopac_step.keyword_metadata
        for keyword in P["extra keywords"]:
            if "=" in keyword:
                keyword, value = keyword.split("=")
                if keyword not in metadata or "format" not in metadata[keyword]:
                    keywords.append(keyword + "=" + value)
                else:
                    keywords.append(metadata[keyword]["format"].format(keyword, value))

        result = []
        result.append([*keywords])

        # Handle MOZYME follow-up calculations
        if "MOZYME" in keywords:
            follow_up = P["MOZYME follow-up"]
            if "exact" in follow_up:
                keywords.remove("MOZYME")
                if "1SCF" not in keywords:
                    keywords.append("1SCF")
                keywords.append("OLDGEO")
                result.append([*keywords])
            elif "new" in follow_up:
                if "1SCF" not in keywords:
                    keywords.append("1SCF")
                keywords.append("OLDGEO")
                result.append([*keywords])
            elif follow_up == "none":
                pass
            else:
                logger.error(f"Don't recognize the MOZYME follow-up: '{follow_up}'")

        return result

    def analyze(self, indent="", data_sections=[], out_sections=[], table=None):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """
        options = self.parent.options

        # Get the parameters used
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        system, starting_configuration = self.get_system_configuration(None)

        if P["MOZYME"] == "always":
            used_mozyme = True
        elif (
            P["MOZYME"] == "for larger systems"
            and starting_configuration.n_atoms >= P["nMOZYME"]
        ):
            used_mozyme = True
        else:
            used_mozyme = False

        if used_mozyme:
            follow_up = P["MOZYME follow-up"]
            if "exact" in follow_up:
                pass
            elif "new" in follow_up:
                pass
            elif follow_up == "none":
                used_mozyme = False
            else:
                logger.error(f"Don't recognize the MOZYME follow-up: '{follow_up}'")

        if table is None:
            table = {
                "Property": [],
                "Value": [],
                "Units": [],
            }
        text = ""

        if used_mozyme:
            data = {**data_sections[0]}
            data.update(data_sections[1])
        else:
            data = data_sections[0]

        if "GRADIENT_NORM" in data:
            tmp = data["GRADIENT_NORM"]
            table["Property"].append("Gradient Norm")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("kcal/mol/Å")

        if "POINT_GROUP" in data and data["POINT_GROUP"] != "":
            text += "The molecule has {POINT_GROUP} symmetry."
        else:
            text += "The symmetry of the molecule was not determined."

        if "HEAT_OF_FORMATION" in data:
            tmp = data["HEAT_OF_FORMATION"]
            table["Property"].append("Enthalpy of Formation")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("kcal/mol")

            tmp = Q_(tmp, "kcal/mol").to("kJ/mol").magnitude
            table["Property"].append("")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("kJ/mol")

            if used_mozyme and "HEAT_OF_FORMATION" in data_sections[0]:
                tmp2 = data_sections[0]["HEAT_OF_FORMATION"]
                tmp2 = Q_(tmp2, "kcal/mol").to("kJ/mol").magnitude
                table["Property"].append("MOZYME Enthalpy of Formation")
                table["Value"].append(f"{tmp2:.2f}")
                table["Units"].append("kJ/mol")

                tmp2 -= tmp
                table["Property"].append("Othonormality error in EoF")
                table["Value"].append(f"{tmp2:.2f}")
                table["Units"].append("kJ/mol")

                if abs(tmp2) > 10.0:
                    text += (
                        "\n\nWarning: The non-orthogonality of the localized orbitals "
                        f"led to an error in the enthalpy of formation of {tmp2:.2f} "
                        "kJ/mol. This is expected if there were many iterations of "
                        "geometry optimization. Otherwise check the results carefully."
                    )
                    if "exact" in follow_up:
                        text += (
                            " You followed up with an exact calculation. A large "
                            "difference in the energy could indicate a problem "
                            "with the localcized molecular orbitals. Check the MOPAC "
                            "output carefully!"
                        )

        if "SPIN_COMPONENT" in data:
            tmp = data["SPIN_COMPONENT"]
            table["Property"].append("Sz")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("")

        if "TOTAL_SPIN" in data:
            tmp = data["TOTAL_SPIN"]
            table["Property"].append("S^2")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("")

        if "IONIZATION_POTENTIAL" in data:
            tmp = data["IONIZATION_POTENTIAL"]
            table["Property"].append("Ionization Energy")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("eV")

        if "DIPOLE" in data:
            tmp = data["DIPOLE"]
            table["Property"].append("Dipole Moment")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("Debye")

        if "EIGENVALUES" in data and "MOLECULAR_ORBITAL_OCCUPANCIES" in data:
            for occ, E in zip(
                data["MOLECULAR_ORBITAL_OCCUPANCIES"], data["EIGENVALUES"]
            ):
                if occ > 0.1:
                    Ehomo = E
                else:
                    Elumo = E
                    break
            table["Property"].append("HOMO Energy")
            table["Value"].append(f"{Ehomo:.2f}")
            table["Units"].append("eV")
            table["Property"].append("LUMO Energy")
            table["Value"].append(f"{Elumo:.2f}")
            table["Units"].append("eV")
            table["Property"].append("Gap")
            table["Value"].append(f"{Elumo - Ehomo:.2f}")
            table["Units"].append("eV")

        if "AREA" in data:
            tmp = data["AREA"]
            table["Property"].append("COSMO Area")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("Å^2")

        if "VOLUME" in data:
            tmp = data["VOLUME"]
            table["Property"].append("COSMO Volume")
            table["Value"].append(f"{tmp:.2f}")
            table["Units"].append("Å^3")

        text_lines = []
        text_lines.append("                     Results")
        text_lines.append(
            tabulate(
                table,
                headers="keys",
                tablefmt="psql",
                colalign=("center", "decimal", "left"),
            )
        )
        text_lines.append("\n\n")

        # Get charges and spins, etc.
        directory = Path(self.directory)
        directory.mkdir(parents=True, exist_ok=True)

        system, configuration = self.get_system_configuration(None)
        symbols = configuration.atoms.symbols
        atoms = configuration.atoms
        if "ATOM_CHARGES" in data:
            # Add to atoms (in coordinate table)
            if "charge" not in atoms:
                atoms.add_attribute(
                    "charge", coltype="float", configuration_dependent=True
                )
            atoms["charge"][0:] = data["ATOM_CHARGES"]

            # Print the charges and dump to a csv file
            chg_tbl = {
                "Atom": [*range(1, len(symbols) + 1)],
                "Element": symbols,
                "Charge": [],
            }
            with open(directory / "atom_properties.csv", "w", newline="") as fd:
                writer = csv.writer(fd)
                if "ATOM_SPINS" in data:
                    header = "        Atomic charges and spins"
                    chg_tbl["Spin"] = []
                    writer.writerow(["Atom", "Element", "Charge", "Spin"])
                    for atom, symbol, q, s in zip(
                        range(1, len(symbols) + 1),
                        symbols,
                        data["ATOM_CHARGES"],
                        data["ATOM_SPINS"],
                    ):
                        q = f"{q:.3f}"
                        s = f"{s:.3f}"

                        writer.writerow([atom, symbol, q, s])

                        chg_tbl["Charge"].append(q)
                        chg_tbl["Spin"].append(s)
                else:
                    header = "        Atomic charges"
                    writer.writerow(["Atom", "Element", "Charge"])
                    for atom, symbol, q in zip(
                        range(1, len(symbols) + 1),
                        symbols,
                        data["ATOM_CHARGES"],
                    ):
                        q = f"{q:.2f}"
                        writer.writerow([atom, symbol, q])

                        chg_tbl["Charge"].append(q)
            if len(symbols) <= int(options["max_atoms_to_print"]):
                text_lines.append(header)
                text_lines.append(
                    tabulate(
                        chg_tbl,
                        headers="keys",
                        tablefmt="psql",
                        colalign=("center", "center"),
                    )
                )
        if "ATOM_SPINS" in data:
            # Add to atoms (in coordinate table)
            if "spin" not in atoms:
                atoms.add_attribute(
                    "spin", coltype="float", configuration_dependent=True
                )
            atoms["spin"][0:] = data["ATOM_SPINS"][0]

        text = str(__(text, **data, indent=self.indent + 4 * " "))
        text += "\n\n"
        text += textwrap.indent("\n".join(text_lines), self.indent + 7 * " ")

        printer.normal(text)

        if used_mozyme and "CPU_TIME" in data_sections[0]:
            t0 = data_sections[0]["CPU_TIME"]
            if "CPU_TIME" in data:
                t1 = data["CPU_TIME"]
                text = (
                    f"This MOZYME calculation took {t0:.2f} s and the follow-up "
                    f"took {t1:.2f} s."
                )
            else:
                text = f"The MOZYME calculation took {t0:.2f} s."
        elif "CPU_TIME" in data:
            t0 = data_sections[0]["CPU_TIME"]
            text = f"This calculation took {t0:.2f} s."
        printer.normal(str(__(text, **data, indent=self.indent + 4 * " ")))

        # Put any requested results into variables or tables
        self.store_results(
            data=data,
            properties=mopac_step.properties,
            results=self.parameters["results"].value,
            create_tables=self.parameters["create tables"].get(),
        )
